import csv
from scipy.special import comb
import numpy as np
from copy import deepcopy


def get_vbs_shap(algorithms, instances, scores, invp = False):
    '''
    instances - the set of different instances that were solved
    algorithms - the set of available algorithms.
    metric - the performance metric from (algorithm,instance) to real number. Higher is better.

    Returns a dictionary from algorithms to their Shapley value, where the coalitional function is
    $$ v(S) = \frac{1}{|X|} \sum_{x\in X} \max_{s\in S} metric(s,x),$$
    where X is the set of instances. This is the "average VBS game" with respect to the given instances,
    algorithms.
    '''
    n = len(algorithms)
    m = len(instances)

    shapleys = {}

    #if true, invert the proportion of the scores so that smaller scores are now greater.
    if invp:
        scores = inverse_proportion(scores)

    d = 0
    #For each instance
    for instance in instances:
        #sys.stderr.write('Processing instance "%s" ...\n' % instance)
        #Sort algorithms by decreasing order of performance.
        #instance_algorithms = sorted([alg for alg in algorithms if re.match("glucose", alg)], key=lambda a : metric(a,instance))
        instance_algorithms = sorted(list(algorithms), key=lambda a :scores[a+instance])


        #print(instance_algorithms)
        d += 1
        #if not(d % len(instances)*.01): print(d/len(instances)*100, "%", "done") #Percentage marks to know the code is processing and not infinite looping

        #For each algorithm, from worst to best.
        for i in range(len(instance_algorithms)):
            #print("loop", instance_algorithms[i])
            ialgorithm = instance_algorithms[i]

            #sys.stderr.write('Processing the rule corresponding to %d-th algorithm "%s" being the best in the coalition.\n' % (i,ialgorithm))

            #The MC-rule is that you must have the algorithm and none of the better ones.
            '''
            If x is the instance and s_1^x,...,s_n^x are sorted from worst, then the rule's pattern is:
            $$ p_i^x = s_i^x \wedge \bigwedge_{j=i+1}^n \overline{s}_j^x $$
            and its weight is 
            $$ w_i^x = metric(s_i^x,x).$$
            '''
            pos = 1
            neg = n-i-1

            #metricvalue = metric(ialgorithm,instance)
            ## normalised as fraction of instances
            #value = 1/float(m)*metricvalue
            value = float(scores[instance_algorithms[i]+instance])
            #sys.stderr.write('Value of this rule : 1/%d, %.4f\n' % (m,value))

            #Calculate the rule Shapley values, and add them to the global Shapley values.

            #Shapley value for positive literals in the rule.
            pos_shap = float(value)/float(pos*comb(pos+neg,neg,exact=True))
            #Shapley value for negative literals in the rule.
            if neg > 0:
                neg_shap = -1*float(value)/float(neg*comb(pos+neg,pos,exact=True))
            else:
                neg_shap = None

            #print(ialgorithm)
            #Update the Shapley value for the literals appearing in the rule.
            for j in range(i,len(instance_algorithms)):
                jalgorithm = instance_algorithms[j]

                if jalgorithm not in shapleys:
                    shapleys[jalgorithm] = 0
                    #print("None")
                if j == i:
                    shapleys[jalgorithm] += pos_shap
                    #print("+",pos_shap, jalgorithm)
                else:
                    shapleys[jalgorithm] += neg_shap
                    #print("-",neg_shap, jalgorithm)
    return shapleys

def get_vbs_shap_temp(algorithms, instances, scores, temporal_order, invp = False):
    '''
    instances - the set of different instances that were solved.
    algorithms - the set of available algorithms.
    temporal_order - the map of version to solvers
    metric - the performance metric from (algorithm,instance) to real number. Higher is better.

    Returns a dictionary from algorithms to their Shapley value, where the coalitional function is
    $$ v(S) = \frac{1}{|X|} \sum_{x\in X} \max_{s\in S} metric(s,x),$$
    where X is the set of instances. This is the "average VBS game" with respect to the given instances,
    algorithms.
    '''

    shapleys = {alg: 0 for alg in algorithms}

    #if true, invert the proportion of the scores so that smaller scores are now greater.
    if invp:
        scores = inverse_proportion(scores)

    n = len(algorithms)

    for instance in instances:
        #sys.stderr.write('Processing instance "%s"...\n' % instance) #progress notification]=

        prevbest = 0

        for version in sorted(temporal_order.keys()):
            #sys.stderr.write('Processing version "%s"...\n' % version)
            # Sort algorithms from that version increasing by metric value
            #instance_algorithms = sorted([alg for alg in algorithms if re.match("glucose", alg)], key=lambda a : metric(a,instance))
            instance_algorithms = sorted(temporal_order[version], key=lambda a : scores[a+instance])
            #sys.stderr.write('Sorted algorithms for %s: %s\n' % (version, instance_algorithms))
            ni = len(instance_algorithms)
        
            for i in range(ni):
                ialgorithm = instance_algorithms[i]

                #sys.stderr.write('Processing the rule corresponding to %d-th algorithm "%s" being the best in the coalition.\n' % (i, ialgorithm))

                # The MC-rule is that you must have the algorithm and none of the better ones.
                '''
                If x is the instance and s_1^x,...,s_n^x are sorted from worst, then the rule's pattern is:
                $$ p_i^x = s_i^x \wedge \bigwedge_{j=i+1}^n \overline{s}_j^x $$
                and its weight is 
                $$ w_i^x = metric(s_i^x,x).$$
                '''
                pos = 1
                neg = ni - i - 1

                if neg < 0:
                    raise Exception("This should never happen.")

                value = max(0, scores[instance_algorithms[i]+instance] - prevbest)
                #sys.stderr.write('Value of this rule : %.4f\n' % (value))

                # Calculate the rule Shapley values, and add them to the global Shapley values.
                combns = float(pos + neg)
                pos_shap = float(value / combns)
                shapleys[ialgorithm] += float(pos_shap)
                
                if neg > 0:
                    neg_shap = value / (neg * combns)
                    for j in range(i + 1, ni):
                        jalgorithm = instance_algorithms[j]
                        shapleys[jalgorithm] -= neg_shap

            prevbest = max(prevbest, scores[instance_algorithms[-1]+instance])

    return shapleys

#Calculates the Shapley value the typical way, as Shapley intended.
#Does not utilize MC-nets for speed up, thus on large sets of data it can take a while to return
#Calculates Shapley value by finding all of the permutations of the set of algorithms
#Than calcuting the average marginal contribution over all of the permutations, this is done for each algorithm
#It returns a dictonary that maps the algorithm to it's shapley score
def traditional_shap(algorithms, instances, scores, invp = False):
    shapleys = {}

    #if true, invert the proportion of the scores so that smaller scores are now greater.
    if invp:
        scores = inverse_proportion(scores)
    
    for algorithm in algorithms: 
        permutations = []
        n = len(algorithms)
        permutate_coalitions(algorithm,algorithms, permutations)

        score = 0
        for perm in permutations:

            with_a = np.sum([max((0 if (scores.get(a+instance) is None) else scores[a+instance]) for a in perm) for instance in instances])
            perm.remove(algorithm)
            
            if not perm:
                without_a = 0
            else:
                without_a = np.sum([max((0 if (scores.get(a+instance) is None) else scores[a+instance]) for a in perm) for instance in instances])

            
            #print(perm, with_a, " - ", without_a)
            score += float(with_a - without_a)/float(comb(n-1,len(perm),exact=True))
        
        score = score/len(algorithms)
        shapleys[algorithm] = score
    return shapleys

#calcultes the marginal contributions for each algorithm
#can be overloaded to marginal and temporal maginal contributions
def marginal_contributions(algorithms, instances, scores, temp_order = None, temp_order_bysolver = None, temp_marges = None, invp = False):
    marges = {}

    #if true, invert the proportion of the scores so that smaller scores are now greater.
    if invp:
        scores = inverse_proportion(scores)

    for algorithm in algorithms:
        leftover_algorithms = list(algorithms)
        leftover_algorithms.remove(algorithm)

        all_all_perf = np.sum([max(scores[a+instance] for a in algorithms) for instance in instances])

        if len(leftover_algorithms) == 0:
            marginal_perf = all_all_perf
        else:
            marginal_perf = all_all_perf - np.sum([max(scores[a+instance] for a in leftover_algorithms) for instance in instances])
        
        marges[algorithm] = marginal_perf

        if temp_marges is not None:
            thisversion = temp_order_bysolver[algorithm]
            versions = [y for y in temp_order.keys() if y <= thisversion]
            available_algorithms = [alg for version in versions for alg in temp_order[version]]

            other_algorithms = list(available_algorithms)
            other_algorithms.remove(algorithm)

            all_perf =  np.sum([max(scores[a+instance] for a in available_algorithms) for instance in instances])
            if len(other_algorithms) == 0:
                tmp_marginal_perf = all_perf
            else:
                tmp_marginal_perf = all_perf - np.sum([max(scores[a+instance] for a in other_algorithms) for instance in instances])
        
            temp_marges[algorithm] = tmp_marginal_perf
    return marges

#calculates the temporal marginal contributions for each algorithm
#just calculates the temporal marginal contributions
def temporal_marginal_contributions(algorithms, instances, scores, temp_order, temp_order_bysolver, invp = False):

    #if true, invert the proportion of the scores so that smaller scores are now greater.
    if invp:
        scores = inverse_proportion(scores)        
    temp_marges = {}
    for algorithm in algorithms:
        thisversion = temp_order_bysolver[algorithm]
        versions = [y for y in temp_order.keys() if y <= thisversion]
        available_algorithms = [alg for version in versions for alg in temp_order[version]]

        other_algorithms = list(available_algorithms)
        other_algorithms.remove(algorithm)

        all_perf = np.sum([max(scores[a+instance] for a in available_algorithms) for instance in instances])
        if len(other_algorithms) == 0:
            tmp_marginal_perf = all_perf
        else:
            tmp_marginal_perf = all_perf - np.sum([max(scores[a+instance] for a in other_algorithms) for instance in instances])
        
        temp_marges[algorithm] = tmp_marginal_perf
    return temp_marges

#Opens the file inputed through the terminal, converts the data in the CVS to information containers the function needs, a
#Algorithms --> gathers all the names of the different algorithms that will be evaluated.
#Instances -->  all the different instances algorithms were run in. 
#Scores --> a dictionary that maps an algorithm used and an instance to the score obtained by that algorithm in that test intsance.
#returns an array of each list in the order, [algortihms, instances, scores]
def read_file(file_name):
    scores = {}
    algorithms = set()
    instances = set()

    with open(file_name) as file_obj:
        top = next(file_obj).replace('\n','')
        header = (top.split (",")) #uses the header to establish what each column is
        a = header.index("algorithm")
        i = header.index("instance")
        p = header.index("performance")
        #print(header)

        data = csv.reader(file_obj)

        for n, row in enumerate(data):
            if row[a] == "" or row[i] == "" or row[p] == "":
                raise ValueError("Missing Data Entry in Row: %d" % (n+2))
            
            algorithm = row[a]
            instance = row[i]
            score = float(row[p])

            algorithms.add(algorithm)
            instances.add(instance)
            if scores.get(algorithm+instance) is None:
                scores[algorithm+instance] = score
            else:
                raise ValueError("duplicate entries, Start at Row: %d" % (n+2))

    return [list(algorithms), list(instances), scores]

#Opens the file inputed through the terminal, converts the data in the CVS to information containers the function needs, a
#This reads the temporal file which provides information about the time an algorithm is made
#This time is used to attribute more or less credit to an algorithm for how much it contributed at that time
#temp_order by solver: The algorithm maps to a time (version)
#temp_order: time(version) maps to to algorithm
def read_temporal_file(file_name, algorithms):
    temp_order = {}
    temp_order_bysolver = {}

    with open(file_name, 'r') as f:
        reader = csv.DictReader(f)
        #print(reader)
        for row in reader:
            if not row['version'] in temp_order:
                temp_order[row['version']] = []
            temp_order[row['version']].append(row['solver'])
            temp_order_bysolver[row['solver']] = row['version']
    for algorithm in algorithms:
        if not algorithm in temp_order_bysolver:
            raise ValueError("No temporal information found for %s!" % algorithm)
    return [temp_order, temp_order_bysolver]

#permutes all the possible subsets of algorithms that includes the inputed algorithm
#Can also be interpreted as generating all the joining orders up to the inputed algorithm of coalitions
def permutate_coalitions(algorithm, left, results):
        if not (left in results):
            results.append(deepcopy(left))

        if len(left) == 1:
            return results
        
        #leftover = [a for a in left if a != algorithm]
        for a in left: 
            if a != algorithm:
                leftover = deepcopy(left)
                leftover.remove(a)
                permutate_coalitions(algorithm, leftover, results)
        
#Returns a new scores dictinary where those with lower scores now have higher scores
#This function is used to translate scores in which lower values are considred better, such as time to solve
def inverse_proportion(scores):
    newScores = deepcopy(scores)
    maxScore = max(scores, key = scores.get)
    
    for key in scores:
        newScores[key] = scores[maxScore] - scores[key]
    
    return newScores



#Functions found to be useful for creating or manipulating temporal dictionaries

#converts a tempOrder dict to a tempOrderbySolver
def toBySolver(tempOrder):
    bySolver = {}
    for time in tempOrder.keys():
        for alg in tempOrder[time]:
            bySolver[alg] = time
    return bySolver

#converts an array of algorithm names to a temporal dictionary, considering the algorithms from the begining to be more recent than those at the end
def toTempOrder(tOrder):
    tempOrder = {}
    for i,algs in enumerate(tOrder):
        tempOrder[str(len(tOrder)-i)] = algs
    return tempOrder


#Code to verify the code of mc_shapley and ensure the calculated MC Shapley value is correct.
from .. import temporal_utils
from .. import mc_shapley as shapley_cvs
import pytest as test

#verifies that a dictionary has the same values according to input array
def dictScores(algorithms, inputScores):
    scores = {}
    for i, alg in enumerate(algorithms):
            scores[alg] = inputScores[i]
    return scores

'''
Function that creates a unit test from array inputs, utlizes only two instances

#ascore = array of scores of alogrithm at an instance, ordered by algorithm than instance, I.E. A1T2 comes before A2T1, A1T1 comes before A1T2
#mscores = array of what marginal contributions of each algorithm should be, algorithms correspond with the index-1
#tmscores = array of what the temporal marginal contributions of algorithms should be, ordered like mscores
#sscores = array of what shapley values of algorithms should be, ordered like mscores
#tscores = array of what temporal shapley values of algorithms should be, ordered like mscores
#tOrder = array of the temporal location of algorithms, ordered like mscores
'''
def unitTestTemplate(ascores, mscores, tmscores, sscores, tsscores, tOrder):
    algorithms = ["A1","A2","A3"]
    instances = ["T-1", "T-2"]


    scores = {"A1T-1": ascores[0],"A1T-2": ascores[1], "A2T-1" : ascores[2], "A2T-2": ascores[3], "A3T-1": ascores[4],  "A3T-2" : ascores[5]}
    tempOrder = temporal_utils.toTempOrder(tOrder)
    tempOrderBySolver = temporal_utils.toBySolver(tempOrder)
    #print(tempOrder, tempOrderBySolver)

    #margAnswersB = [0.29,0,1.31]

    margAnswers = {"A1": mscores[0], "A2": mscores[1], "A3" :mscores[2]}
    tempMargAnswers = {"A1": tmscores[0], "A2": tmscores[1], "A3" : tmscores[2]}
    tempShapAnswers = {"A1": tsscores[0], "A2": tsscores[1], "A3" : tsscores[2]}
    shapAnswers ={"A1": sscores[0], "A2": sscores[1], "A3" : sscores[2]}
    



    #answers = [margAnswers, tempMargAnswers, tempShapAnswers, shapAnswers]

    marg = shapley_cvs.marginal_contributions(algorithms, instances, scores) 
    tempMarg = shapley_cvs.temporal_marginal_contributions(algorithms, instances, scores, tempOrder, tempOrderBySolver)
    shap = shapley_cvs.get_vbs_shap(algorithms, instances, scores)
    tempShap = shapley_cvs.get_vbs_shap_temp(algorithms, instances, scores, tempOrder)

    tradShap = shapley_cvs.traditional_shap(algorithms, instances, scores)
    #print(tempMarg)

    #print(dictScores(frozenset(algorithms),margAnswersB), marg)

    functions = [(marg, margAnswers, "Marginal Contribution"), (tempMarg, tempMargAnswers, "Temporal Marginal Contribution"), 
                 (shap, shapAnswers, "Shapley Value"), (tempShap, tempShapAnswers,"Temporal Shapley"), 
                 (tradShap, shapAnswers, "Traditional Shapley Value")]
    for func in functions:
        #print(func[0], func[1])
        assert func[0] == test.approx(func[1], rel=1e-4), func[2] + " UNIT TEST FAILURE: \nReturned: " + str(func[0]) + " does not equal answer:" + str(func[1]) + "\n"


def test_unit_test_1():
    unitTestTemplate(ascores=[1.02,0,0.73,0,0,1], mscores=[0.29,0,1], tmscores=[0.29, 0, 1], sscores=[0.655,0.365,1], tsscores=[0.655,0.365,1], tOrder=[["A1","A2","A3"]])

def test_unit_test_2():
    unitTestTemplate(ascores=[1.02,0,0.73,0,0,1.31], mscores=[0.29,0, 1.31], tmscores=[0.29, 0.73, 1.31], sscores=[0.655,0.365,1.31], tsscores=[0.29,0.73,1.31], tOrder=[["A3"],["A1"],["A2"]])

def test_unit_test_3():
    unitTestTemplate(ascores=[0.5,0.5,1,0,0,1.0], mscores=[0,0.5,0.5], tmscores=[1,0.5,0.5], sscores=[0.5,0.75,0.75], tsscores=[1,0.5,0.5], tOrder=[["A3","A2"],["A1"]])

def test_unit_test_4():
    unitTestTemplate(ascores=[1,1,1,1,1,1],mscores=[0,0,0],tmscores=[0,0,2], sscores=[0.66666,0.66666,0.66666], tsscores=[0,0,2], tOrder=[["A1","A2"],["A3"]])

def test_unit_test_5():
    unitTestTemplate(ascores=[1.02,0,0.73,0,0,0], mscores=[0.29,0,0], tmscores=[0.29, 0, 0], sscores=[0.655,0.365,0], tsscores=[0.655,0.365,0], tOrder=[["A1","A2"], ["A3"]])

def test_traditional():
    [algorithms, instances, scores] = shapley_cvs.read_file("./tests_mc_shapley/test_file2.txt") 
    assert shapley_cvs.traditional_shap(algorithms, instances, scores) == test.approx(shapley_cvs.get_vbs_shap(algorithms, instances, scores), rel=1e-4)

def test_better():
    scores = {'a ': 1, 'b' : 2, 'c' : 3, 'd' : 4, 'e' : 5, 'f' : 10}
    newScores = shapley_cvs.inverse_proportion(scores)
    assert newScores == {'a ': 9, 'b' : 8, 'c' : 7, 'd' : 6, 'e' : 5, 'f' : 0}

def test_flag():
    [algorithms, instances, scores] = shapley_cvs.read_file("./tests_mc_shapley/test_file2.txt") 
    newScores = shapley_cvs.inverse_proportion(scores)
    assert shapley_cvs.get_vbs_shap(algorithms, instances, newScores) == test.approx(shapley_cvs.get_vbs_shap(algorithms, instances, scores, invp=True), rel=1e-4)

def test_readfile_1(): 
        algorithms = set(["A1", "A2", "A3"])
        instances = set(["1-1", "1-2"])
        scores = {"A11-1": 1.02, "A21-1": 0.73, "A31-1": 0.3, "A11-2": 1.0, "A21-2": 0.5, "A31-2": 0.21}

        ais = shapley_cvs.read_file("./tests_mc_shapley/test_file1.txt")
        assert set(ais[0]) == algorithms, "Algorithms were not read properly: " + str(ais[0]) + " does not equal " + str(algorithms)
        assert set(ais[1]) == instances, "Instances were not read in properly: " + str(ais[1]) + " does not equal " + str(instances)
        assert ais[2] == scores, 'Score dictionary is malformed: ' + str(ais[2]) + " does not equal " + str(scores)

def test_readfile_2(): 
        algorithms = set(["insertion", "first", "random"])
        instances = set(["1-10000", "2-10000"])
        scores = {"insertion1-10000": 0, "first1-10000": 0, "random1-10000": 0.01, "insertion2-10000": 0, "first2-10000": 0.96, "random2-10000": 0.01}

        ais = shapley_cvs.read_file("./tests_mc_shapley/test_file2.txt")
        assert set(ais[0]) == algorithms, "Algorithms were not read properly: " + str(ais[0]) + " does not equal " + str(algorithms) + "\n"
        assert set(ais[1]) == instances, "Instances were not read in properly: " + str(ais[1]) + " does not equal " + str(instances) + "\n"
        assert ais[2] == scores, 'Score dictionary is malformed: ' + str(ais[2]) + " does not equal " + str(scores) + "\n"

        t_order = {"1946": ["insertion"], "1961": ["first", "random"]}
        t_order_bysolver = {"insertion": "1946", "first": "1961", "random": "1961"}
        temporal = shapley_cvs.read_temporal_file("./tests_mc_shapley/temp_test_file2.csv", ais[0])
        assert temporal[0] == t_order, "Temporal order was not read properly: " + str(temporal[0]) + " does not equal " + str(t_order) + "\n" 
        assert temporal[1] == t_order_bysolver, "Temporal order by solver was not formed properly: " + str(temporal[1]) + " does not equal " + str(t_order_bysolver) + "\n" 

def test_readfile_3(): 
    algorithms = set(["A1", "A2", "A3", "A4", "A5"])
    instances = set(["1-1", "1-2", "1-3", "1-4", "1-5", "1-6", "1-7"])

    ais = shapley_cvs.read_file("./tests_mc_shapley/test_file3.csv")
    assert set(ais[0]) == algorithms, "Algorithms were not read properly: " + str(ais[0]) + " does not equal " + str(algorithms) + "\n"
    assert set(ais[1]) == instances, "Instances were not read in properly: " + str(ais[1]) + " does not equal " + str(instances) + "\n"
    assert ais[2]["A21-3"] == 0.11, "Score dictionary is malformed: " + str(ais[2]["A21-3"]) + " does not equal 0.11" + "\n"
    assert ais[2]["A11-3"] == 0.15, "Score dictionary is malformed: " + str(ais[2]["A11-3"]) + " does not equal 0.15" + "\n"
    assert ais[2]["A41-3"] == 0.45, "Score dictionary is malformed: " + str(ais[2]["A41-3"]) + " does not equal 0.45" + "\n"
    assert ais[2]["A41-5"] == 0.28, "Score dictionary is malformed: " + str(ais[2]["A41-5"]) + " does not equal 0.28" + "\n"
    assert ais[2]["A51-7"] == 1, "Score dictionary is malformed: " + str(ais[2]["A51-7"]) + " does not equal 0.15" + "\n"
    assert ais[2].get("A31-4") is  None, "Failure, wrote new information into data: A3, 1-4"

def test_missing_entry(): 
    with test.raises(ValueError) as exc_info: 
        shapley_cvs.read_file("./tests_mc_shapley/test_file4.csv")
    assert str(exc_info.value) == "Missing Data Entry in Row: 18", "Failed to raise proper exception: " 

def test_score_conflict():
    with test.raises(ValueError) as exc_info: 
        shapley_cvs.read_file("./tests_mc_shapley/test_file5.csv")
    assert str(exc_info.value) == "duplicate entries, Start at Row: 17", "Failed to raise proper exception: " 

def test_temporal_missing(): 
    algorithms = ["A1", "A2", "A3", "A4", "A5"]
    with test.raises(ValueError) as exc_info: 
        shapley_cvs.read_temporal_file("./tests_mc_shapley/temp_test_file3.csv", algorithms)
    assert str(exc_info.value) == "No temporal information found for A5!", "Failed to raise proper exception: " 

def test_kotthoff():
    [algorithms, instances, scores] = shapley_cvs.read_file("./tests_mc_shapley/test_file_park.csv") 
    shapleys = shapley_cvs.get_vbs_shap(algorithms, instances, scores, invp=False)
    print(shapleys)
    assert shapleys == test.approx({'qsort4': 0, '1978 + cutoff + median 3': 0, '1993 + median 9 + split end': 0, '1993 + median 9': 0, '2009': 0}, rel=1e-4)